"""Generate features over a dataframe."""

import warnings

import numpy as np
import pandas as pd
import tqdm
from feature_engine.datetime import DatetimeFeatures


def _ticker_features(df: pd.DataFrame, windows: list[int]) -> pd.DataFrame:
    cols = df.columns.values.tolist()
    for col in tqdm.tqdm(cols, desc="Generating ticker features"):
        s = df[col]
        for w in windows:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=pd.errors.PerformanceWarning)
                # SMA
                sma = s.rolling(w).mean()
                df[f"{col}_sma_{w}"] = sma / s - 1
                # PCT
                df[f"{col}_pctchg_{w}"] = s.pct_change(w, fill_method=None)
                # Z-Score
                mu = s.rolling(w).mean()
                sigma = s.rolling(w).std()
                df[f"{col}_z_{w}"] = (s - mu) / sigma
    return df


def _meta_ticker_feature(
    df: pd.DataFrame, lags: list[int], windows: list[int]
) -> pd.DataFrame:
    dfs = [df]
    for lag in tqdm.tqdm(lags, desc="Generating lags"):
        dfs.append(df.shift(lag).add_suffix(f"_lag{lag}"))
    for window in tqdm.tqdm(windows, desc="Generating window features"):
        dfs.append(df.rolling(window).mean().add_suffix(f"_rmean{window}"))  # type: ignore
        dfs.append(df.rolling(window).std().add_suffix(f"_rstd{window}"))  # type: ignore
    return pd.concat(dfs, axis=1).replace([np.inf, -np.inf], np.nan)


def _dt_features(df: pd.DataFrame) -> pd.DataFrame:
    print("Generating datetime features")
    dtf = DatetimeFeatures(features_to_extract="all", variables="index")
    return dtf.fit_transform(df)


def features(df: pd.DataFrame, windows: list[int], lags: list[int]) -> pd.DataFrame:
    """Generate features on a dataframe."""
    cols = df.columns.values.tolist()
    df = _ticker_features(df=df, windows=windows)
    df = _meta_ticker_feature(df, lags=lags, windows=windows)
    df = _dt_features(df=df)
    return df.drop(columns=cols)



"""The CLI for finding mispriced options."""

import argparse
import datetime

import requests_cache
import tqdm
import wavetrainer as wt

from .download import download
from .features import features
from .normalizer import denormalize, normalize

_TICKERS = [
    # Equities
    "SPY",
    "QQQ",
    "EEM",
    # Commodities
    "GC=F",
    "CL=F",
    "SI=F",
    # FX
    "EURUSD=X",
    "USDJPY=X",
    # Crypto
    "BTC-USD",
    "ETH-USD",
]
_MACROS = [
    "GDP",
    "UNRATE",
    "CPIAUCSL",
    "FEDFUNDS",
    "DGS10",
    "T10Y2Y",
    "M2SL",
    "VIXCLS",
    "DTWEXBGS",
    "INDPRO",
]
_WINDOWS = [
    5,
    10,
    20,
    60,
    120,
    200,
]
_LAGS = [1, 3, 5, 10, 20, 30]
_DAYS_OUT = 30


def main() -> None:
    """The main CLI function."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--inference",
        help="Whether to skip training and just do inference.",
        required=False,
        default=False,
        action="store_true",
    )
    args = parser.parse_args()

    # Setup main objects
    session = requests_cache.CachedSession("panelbeater-cache")
    wavetrainer = wt.create(
        "panelbeater-train",
        walkforward_timedelta=datetime.timedelta(days=30),
        validation_size=datetime.timedelta(days=365),
        test_size=datetime.timedelta(days=365),
        allowed_models={"catboost"},
        max_false_positive_reduction_steps=0,
    )

    # Fit the models
    df_y = download(tickers=_TICKERS, macros=_MACROS, session=session)
    df_x = features(df=df_y.copy(), windows=_WINDOWS, lags=_LAGS)
    df_y_norm = normalize(df=df_y.copy())
    if not args.inference:
        wavetrainer.fit(df_x, y=df_y_norm)
    for _ in tqdm.tqdm(range(_DAYS_OUT), desc="Running t+X simulation"):
        df_next = wavetrainer.transform(df_x, ignore_no_dates=True).drop(columns=df_x)
        df_y = denormalize(df_next)
        df_x = features(df=df_y.copy(), windows=_WINDOWS, lags=_LAGS)
        df_y_norm = normalize(df=df_y.copy())

    # Find the current options prices


if __name__ == "__main__":
    main()
